% -*- MATLAB -*-
% Time-stamp: "2018-08-12 10:14:30 fujiwara"
%----------------------------------------------------------------------
%  exfloat class for Octave 4
%  Copyright (C) 2018, FUJIWARA,Hiroshi
%    <fujiwara@acs.i.kyoto-u.ac.jp>
%----------------------------------------------------------------------

classdef exfloat
   
  properties( Constant, ACCESS=private )

    %----------------------------------------
    % required precision in decimal digits
    %----------------------------------------
    precision10 = 100;

    % default output format
    default_output_precision = int32(4);
    default_output_format    = 'd';         % 'd' means 'as double'
    default_display_format   = '%.20g';
    default_fixed_scientific = -4;

    SIGN     = uint64( hex2dec('8000000000000000') );
    NEGINFTY = uint64( hex2dec('ffffffffffffffff') );
    F_FORMAT = uint64( hex2dec('0400000000000000') );
    E_FORMAT = uint64( hex2dec('0200000000000000') );
    EXP_MASK = bitshift(uint64(hex2dec('7fffffff')), 32)+uint64(hex2dec('ffffffff'));
    BIAS     = bitshift(uint64(hex2dec('3fffffff')), 32)+uint64(hex2dec('ffffffff'));

  endproperties

  %----------------------------------------
  % MATLAB accepts following initializations
  % in properties( Constant )  block
  %----------------------------------------
%%%    % precision in 2^64 radix, log(10)/log(2)/64 = 0.051905126
%%%    precision64 = int32(exfloat.precision10 * 0.051905126) + 1;
%%%
%%%    % actual arithmetic precision, 64*log(2) = 19.26591972
%%%    precision = 19.26591972 * double(exfloat.precision64);
%%%
%%%    % sum of exponent part and fraction part
%%%    size64 = int32(1 + exfloat.precision64);
%%%    byte   = int32(8 * exfloat.size64);
  
  %----------------------------------------
  % alternatives in Octave
  %----------------------------------------
  methods (Static)

    % precision in 2^64 radix, log(10)/log(2)/64 = 0.051905126
    function p = precision64()
      p = int32(exfloat.precision10 * 0.051905126) + 1;
    endfunction

    % actual arithmetic precision, 64*log(2) = 19.26591972
    function p = precision()
      p = 19.26591972 * double(exfloat.precision64());
    endfunction

    % sum of exponent part and fraction part
    function p = size64()
      p = int32(1 + exfloat.precision64());
    endfunction

    function p = byte()
      p = int32(8 * exfloat.size64());
    endfunction

  endmethods

  properties( ACCESS=public )
    num
  endproperties

  methods
  
    %------------------------------
    % constructor
    %------------------------------
    function x = exfloat( a )

      if nargin == 0
        x.num = zeros(exfloat.size64(),1, 'uint64');
        return;
      endif
      
      [row,col] = size( a );

      if isa(a, 'exfloat')

        x.num = a.num;

      elseif isa(a, 'integer')

        x.num = zeros(exfloat.size64(),1, 'uint64');
        exflib_float_ltoef(x.num, a, exfloat.precision64());

      elseif isa(a, 'char')

        x.num = zeros(exfloat.size64(),1, 'uint64');
        exflib_float_etoef(a, x.num, exfloat.precision64());

      elseif (isa(a,'double') || isa(a,'float')) ...
            && (a-fix(a) == 0)  % essentially integer

        x.num = zeros(exfloat.size64(),1, 'uint64');
        exflib_float_ltoef(x.num, int64(a), exfloat.precision64());

      elseif ((row~=1) || (col~=1))

        error('conversion from array not yet implemented');
        return;

      else

        error('Implicit convertion from %d is prohibited.', a);
        return;

      endif

    endfunction % x=exfloat(a)

    %%% tentative
    function disp( x )
        printf("disp\n");
    endfunction

  endmethods

  methods

    function cast( x, d )
    
      % explicit conversion from double to exfloat
      % usage:
      %    f = exfloat();
      %    f.cast(0.1);
      
      if isa(d, 'double')
        exflib_float_dtoef(d, x.num, exfloat.precision64());
      else
        error('exflib_cast: conversion from unknown type');
      endif
      
    endfunction % cast

  endmethods

  methods (Static)

    %------------------------------
    % auxiliary routines
    %------------------------------
    function p = get_precision()
      p = exfloat.precision();
    endfunction

    function p = get_req_precision()
      p = exfloat.precision10;
    endfunction

    function p = get_exfloat_precision64()
      p = exfloat.precision64();
    endfunction

    function p = get_exfloat_byte()
      p = exfloat.byte();
    endfunction

  endmethods % (Static)

  methods (Static, Access=private)

    function [ output_prec, output_fmt ] = parse_fmt ( fmt )

      % default
      output_prec = exfloat.default_output_precision;
      output_fmt  = exfloat.default_output_format; % default

      if ~isa(fmt, 'char')
        %warning('fmt should be integer : %s', fmt) 
       return
      endif

      if ( fmt(1)=='%' && (fmt(2)=='f' || fmt(2)=='F') )
        output_fmt = 'f'; % fixed format
        return
      endif

      if ( fmt(1)=='%' && (fmt(2)=='e' || fmt(2)=='E') )
        output_fmt = 'e'; % scientific format
        return
      endif

      if ( fmt(1)=='%' && (fmt(2)=='g' || fmt(2)=='G') )
        output_fmt = 'g'; % generic format';
        return
      endif
	
      if ~( fmt(1)=='%' && fmt(2)=='.' )
        %warning('illeagal format : %s', fmt)
        return
      endif

      output_prec = 0;
      for n=3:length(fmt)

        d = fmt(n) - '0';

        if (d < 0) || (9 < d)
          break
        endif
    
        output_prec = output_prec * 10;
        output_prec = output_prec + d;
      endfor

      if fmt(n) == 'e' || fmt(n) == 'E'
        output_fmt = 'e'; % scientific format
      elseif fmt(n) == 'f' || fmt(n) == 'F'
        output_fmt = 'f'; % fixed format
      elseif fmt(n) == 'g' || fmt(n) == 'G'
        output_fmt = 'g'; % generic format';
      else
        %warning('unknown format : %s', fmt)
      endif

    endfunction % [output_prec,output_fmt]=parse_fmt(fmt)

  endmethods % methods (Access=private)

endclassdef

%----------------------------------------------------------------------
% End of file
%----------------------------------------------------------------------
